from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Entity, EntityRole, EntityType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestEntityTypesAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.location_type = EntityType.objects.get(name='location')
        cls.date_type = EntityType.objects.get(name='date')
        cls.org_type = EntityType.objects.get(name='organization')
        cls.person_type = EntityType.objects.get(name='person')
        cls.number_type = EntityType.objects.get(name='number')
        cls.unverified_user = User.objects.create(email='user@address.com', password='P4$5w0Rd')
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.private_corpus_admin = User.objects.create(email='jake@seresin.com', password='h_ngm_n', verified_email=True)
        cls.private_corpus_guest = User.objects.create(email='pete@mitchell.com', password='maverick', verified_email=True)
        cls.no_access = User.objects.create(email='bradley@bradshaw.com', password='rooster', verified_email=True)
        # Admin user on private_corpus
        cls.private_corpus.memberships.create(user=cls.private_corpus_admin, level=Role.Admin.value)
        # Contributor user on private_corpus
        cls.private_corpus.memberships.create(user=cls.user, level=Role.Contributor.value)
        # Guest user on private_corpus
        cls.private_corpus.memberships.create(user=cls.private_corpus_guest, level=Role.Guest.value)
        cls.private_corpus_entity_type = EntityType.objects.create(name='some type', corpus=cls.private_corpus)

    # ENTITY TYPE CREATE

    def test_create_entity_type_requires_login(self):
        data = {
            'corpus': str(self.corpus.id),
            'name': 'dagger one'
        }
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_entity_type_requires_verified(self):
        self.client.force_login(self.unverified_user)
        data = {
            'corpus': str(self.corpus.id),
            'name': 'dagger one'
        }
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have permission to perform this action.'}
        )

    def test_create_entity_type_requires_corpus_admin(self):
        self.client.force_login(self.user)
        data = {
            'corpus': str(self.private_corpus.id),
            'name': 'dagger one'
        }
        self.assertEqual(self.private_corpus.entity_types.all().count(), 1)
        with self.assertNumQueries(5):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'corpus': [f'Invalid pk "{str(self.private_corpus.id)}" - object does not exist.']}
        )
        self.assertEqual(self.private_corpus.entity_types.all().count(), 1)

    def test_create_entity_type(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'corpus': str(self.private_corpus.id),
            'name': 'dagger spare'
        }
        self.assertEqual(self.private_corpus.entity_types.all().count(), 1)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.private_corpus.entity_types.all().count(), 2)
        self.assertEqual(response.json()['color'], 'ff0000')
        self.assertEqual(response.json()['name'], 'dagger spare')
        self.assertEqual(response.json()['corpus'], str(self.private_corpus.id))

    def test_create_entity_type_color_validation(self):
        self.client.force_login(self.private_corpus_admin)
        cases = [
            [
                'phoenix',
                [
                    'Color should be a hexadecimal color code',
                    'Ensure this field has no more than 6 characters.'
                ]
            ],
            [
                'yellow',
                ['Color should be a hexadecimal color code']
            ],
            [
                'YELLOW',
                ['Color should be a hexadecimal color code']
            ]
        ]
        for color, errors in cases:
            with (self.subTest(color=color, errors=errors), self.assertNumQueries(5)):
                response = self.client.post(
                    reverse('api:entity-type-create'),
                    data={
                        'corpus': str(self.private_corpus.id),
                        'name': 'dagger spare',
                        'color': color
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), {'color': errors})

    def test_create_entity_type_color_ignore_case(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'corpus': str(self.private_corpus.id),
            'name': 'dagger spare',
            'color': 'FF0000'
        }
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['color'], 'FF0000')

    def test_create_entity_type_color(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'corpus': str(self.private_corpus.id),
            'name': 'dagger spare',
            'color': 'f76719'
        }
        self.assertEqual(self.private_corpus.entity_types.all().count(), 1)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.private_corpus.entity_types.all().count(), 2)
        self.assertEqual(response.json()['color'], 'f76719')
        self.assertEqual(response.json()['name'], 'dagger spare')
        self.assertEqual(response.json()['corpus'], str(self.private_corpus.id))

    def test_create_entity_type_already_exists(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'corpus': str(self.private_corpus.id),
            'name': 'some type',
            'color': 'f76719'
        }
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:entity-type-create'), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['Entity Type with name some type already exists in corpus private.']})

    # ENTITY TYPE UPDATE

    def test_update_entity_type_requires_login(self):
        data = {
            'name': 'dagger one',
            'color': 'f76719'
        }
        with self.assertNumQueries(0):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_entity_type_requires_verified(self):
        self.client.force_login(self.unverified_user)
        data = {
            'name': 'dagger one',
            'color': 'f76719'
        }
        with self.assertNumQueries(2):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have permission to perform this action.'}
        )

    def test_update_entity_type_requires_corpus_admin(self):
        self.client.force_login(self.user)
        data = {
            'name': 'dagger one',
            'color': 'f76719'
        }
        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this corpus.'}
        )
        self.private_corpus_entity_type.refresh_from_db()
        self.assertEqual(self.private_corpus_entity_type.name, 'some type')
        self.assertEqual(self.private_corpus_entity_type.color, 'ff0000')

    def test_update_entity_type(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'name': 'dagger spare',
            'color': 'f76719'
        }
        with self.assertNumQueries(9):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.private_corpus_entity_type.refresh_from_db()
        self.assertEqual(self.private_corpus_entity_type.color, 'f76719')
        self.assertEqual(self.private_corpus_entity_type.name, 'dagger spare')

    def test_update_entity_type_name_already_exists(self):
        self.client.force_login(self.private_corpus_admin)
        EntityType.objects.create(corpus=self.private_corpus, name='dagger spare')
        data = {
            'name': 'dagger spare',
            'color': 'f76719'
        }
        with self.assertNumQueries(8):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['Entity Type with name dagger spare already exists in corpus private.']})

    def test_update_entity_requires_name(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'color': 'f76719'
        }
        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['This field is required.']})

    def test_update_entity_no_corpus_change(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'corpus': str(self.corpus.id),
            'name': 'iceman'
        }
        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'corpus': ["It is not possible to update an Entity Type's corpus."]})

    # ENTITY TYPE PARTIAL UPDATE

    def test_partial_update_entity_type_requires_login(self):
        data = {
            'name': 'dagger one',
            'color': 'f76719'
        }
        with self.assertNumQueries(0):
            response = self.client.patch(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_entity_type_requires_verified(self):
        self.client.force_login(self.unverified_user)
        data = {
            'name': 'dagger one',
            'color': 'f76719'
        }
        with self.assertNumQueries(2):
            response = self.client.patch(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have permission to perform this action.'}
        )

    def test_partial_update_entity_type_requires_corpus_admin(self):
        self.client.force_login(self.user)
        data = {
            'name': 'dagger one',
            'color': 'f76719'
        }
        with self.assertNumQueries(7):
            response = self.client.patch(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this corpus.'}
        )
        self.private_corpus_entity_type.refresh_from_db()
        self.assertEqual(self.private_corpus_entity_type.name, 'some type')
        self.assertEqual(self.private_corpus_entity_type.color, 'ff0000')

    def test_partial_update_entity_type(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'name': 'dagger spare',
            'color': 'f76719'
        }
        with self.assertNumQueries(9):
            response = self.client.patch(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['color'], 'f76719')
        self.assertEqual(response.json()['name'], 'dagger spare')

    def test_partial_update_entity_is_partial(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'color': 'f76719'
        }
        with self.assertNumQueries(9):
            response = self.client.patch(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.private_corpus_entity_type.refresh_from_db()
        self.assertEqual(self.private_corpus_entity_type.color, 'f76719')
        self.assertEqual(self.private_corpus_entity_type.name, 'some type')

    def test_partial_update_entity_no_corpus_change(self):
        self.client.force_login(self.private_corpus_admin)
        data = {
            'corpus': str(self.corpus.id),
        }
        with self.assertNumQueries(7):
            response = self.client.patch(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), data=data, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'corpus': ["It is not possible to update an Entity Type's corpus."]})

    # ENTITY TYPE RETRIEVE

    def test_retrieve_entity_type_does_not_require_login_public_project(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:entity-type-details', kwargs={'pk': self.location_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.location_type.id),
            'corpus': str(self.corpus.id),
            'name': 'location',
            'color': 'ff0000'
        })

    def test_retrieve_entity_type_requires_private_corpus_guest(self):
        self.client.force_login(self.no_access)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(
            response.json(),
            {'detail': 'Not found.'}
        )

    def test_retrieve_entity_type(self):
        self.client.force_login(self.private_corpus_guest)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['color'], 'ff0000')
        self.assertEqual(response.json()['name'], 'some type')

    # ENTITY TYPE DELETE

    def test_delete_entity_type_requires_corpus_admin(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this corpus.'}
        )

    def test_delete_entity_type(self):
        self.client.force_login(self.private_corpus_admin)
        with self.assertNumQueries(10):
            response = self.client.delete(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(EntityType.DoesNotExist):
            self.private_corpus_entity_type.refresh_from_db()

    def test_delete_entity_type_has_entities(self):
        Entity.objects.create(type=self.private_corpus_entity_type, corpus=self.private_corpus, name='bob')
        self.client.force_login(self.private_corpus_admin)
        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'detail': ['Some entities are using this entity type.']})

    def test_delete_entity_type_has_roles(self):
        EntityRole.objects.create(
            corpus=self.private_corpus,
            parent_name="goose",
            child_name="rooster",
            parent_type=self.private_corpus_entity_type,
            child_type=self.private_corpus_entity_type
        )
        self.client.force_login(self.private_corpus_admin)
        with self.assertNumQueries(9):
            response = self.client.delete(reverse('api:entity-type-details', kwargs={'pk': self.private_corpus_entity_type.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'detail': ['Some entity roles are using this entity type.']})

    # LIST CORPUS ENTITY TYPES

    def test_corpus_entity_types_does_not_require_login_public_project(self):
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:corpus-entity-types', kwargs={'pk': self.corpus.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json()['results'], [
            {
                'id': str(self.location_type.id),
                'name': 'location',
                'color': 'ff0000'
            },
            {
                'id': str(self.person_type.id),
                'name': 'person',
                'color': 'ff0000'
            },
            {
                'id': str(self.date_type.id),
                'name': 'date',
                'color': 'ff0000'
            },
            {
                'id': str(self.number_type.id),
                'name': 'number',
                'color': 'ff0000'
            },
            {
                'id': str(self.org_type.id),
                'name': 'organization',
                'color': 'ff0000'
            },
        ])

    def test_corpus_entity_types_requires_guest_private_project(self):
        self.client.force_login(self.no_access)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-entity-types', kwargs={'pk': self.private_corpus.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_entity_types(self):
        type_2 = EntityType.objects.create(corpus=self.private_corpus, color='f01ef7', name='type 2')
        type_3 = EntityType.objects.create(corpus=self.private_corpus, color='29d911', name='person')
        self.client.force_login(self.private_corpus_guest)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:corpus-entity-types', kwargs={'pk': self.private_corpus.id}), format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json()['results'], [
            {
                'id': str(type_3.id),
                'name': 'person',
                'color': '29d911'
            },
            {
                'id': str(self.private_corpus_entity_type.id),
                'name': 'some type',
                'color': 'ff0000'
            },
            {
                'id': str(type_2.id),
                'name': 'type 2',
                'color': 'f01ef7'
            }
        ])
