import hashlib
import uuid

from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.authentication import AgentUser
from arkindex.ponos.models import Farm
from arkindex.process.models import Process, ProcessMode, Repository
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_public_key
from arkindex.users.models import Right, Role, User


@override_settings(PONOS_LOG_TAIL=42)
class TestAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.wheat_farm = Farm.objects.get(name='Wheat farm')
        pubkey = build_public_key()
        cls.agent = AgentUser.objects.create(
            id=uuid.UUID(hashlib.md5(pubkey.encode("utf-8")).hexdigest()),
            farm=cls.wheat_farm,
            hostname="ghostname",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=2e9,
            last_ping=timezone.now(),
        )

        cls.repo = Repository.objects.first()
        cls.repository_process = Process.objects.create(
            mode=ProcessMode.Repository,
            creator=cls.superuser,
            revision=cls.repo.revisions.first(),
        )

        # Make corpus private
        cls.corpus.public = False
        cls.corpus.save()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.run()
        # Two tasks from the same process
        cls.task1, cls.task2 = cls.process.tasks.all()[:2]
        # Brand new user and corpus with no preexisting rights
        new_user = User.objects.create(email='another@user.com')
        new_corpus = Corpus.objects.create(name='corpus the second')
        Right.objects.create(user=new_user, content_object=new_corpus, level=50)
        cls.process2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=new_user,
            corpus=new_corpus,
            farm=cls.wheat_farm,
        )
        cls.process2.run()
        cls.task3 = cls.process2.tasks.first()

        cls.artifact1 = cls.task1.artifacts.create(
            path="path/to/file.json",
            content_type="application/json",
            size=42,
        )
        cls.artifact2 = cls.task1.artifacts.create(
            path="some/text.txt",
            content_type="text/plain",
            size=1337,
        )

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:task-artifacts", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:task-artifacts", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_process_guest(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.memberships.filter(user=self.user).delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:task-artifacts", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_process_level_corpus(self):
        self.process.creator = self.superuser
        self.process.save()
        self.client.force_login(self.user)

        for role in Role:
            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)

                with self.assertNumQueries(6):
                    response = self.client.get(reverse("api:task-artifacts", args=[self.task1.id]))
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertListEqual(
                    response.json(),
                    [
                        {
                            "content_type": "application/json",
                            "created": self.artifact1.created.isoformat().replace("+00:00", "Z"),
                            "id": str(self.artifact1.id),
                            "path": "path/to/file.json",
                            "s3_put_url": None,
                            "size": 42,
                            "updated": self.artifact1.updated.isoformat().replace("+00:00", "Z"),
                        },
                        {
                            "content_type": "text/plain",
                            "created": self.artifact2.created.isoformat().replace("+00:00", "Z"),
                            "id": str(self.artifact2.id),
                            "path": "some/text.txt",
                            "s3_put_url": None,
                            "size": 1337,
                            "updated": self.artifact2.updated.isoformat().replace("+00:00", "Z"),
                        },
                    ],
                )

    def test_list_process_level_repo(self):
        self.client.force_login(self.user)
        membership = self.repo.memberships.create(user=self.user, level=Role.Guest.value)

        for role in Role:
            with self.subTest(role=role):
                membership.level = role.value
                membership.save()

                with self.assertNumQueries(7):
                    response = self.client.get(reverse("api:task-artifacts", args=[self.task1.id]))
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertListEqual(
                    response.json(),
                    [
                        {
                            "content_type": "application/json",
                            "created": self.artifact1.created.isoformat().replace("+00:00", "Z"),
                            "id": str(self.artifact1.id),
                            "path": "path/to/file.json",
                            "s3_put_url": None,
                            "size": 42,
                            "updated": self.artifact1.updated.isoformat().replace("+00:00", "Z"),
                        },
                        {
                            "content_type": "text/plain",
                            "created": self.artifact2.created.isoformat().replace("+00:00", "Z"),
                            "id": str(self.artifact2.id),
                            "path": "some/text.txt",
                            "s3_put_url": None,
                            "size": 1337,
                            "updated": self.artifact2.updated.isoformat().replace("+00:00", "Z"),
                        },
                    ],
                )

    def test_list_admin(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:task-artifacts", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    "content_type": "application/json",
                    "created": self.artifact1.created.isoformat().replace("+00:00", "Z"),
                    "id": str(self.artifact1.id),
                    "path": "path/to/file.json",
                    "s3_put_url": None,
                    "size": 42,
                    "updated": self.artifact1.updated.isoformat().replace("+00:00", "Z"),
                },
                {
                    "content_type": "text/plain",
                    "created": self.artifact2.created.isoformat().replace("+00:00", "Z"),
                    "id": str(self.artifact2.id),
                    "path": "some/text.txt",
                    "s3_put_url": None,
                    "size": 1337,
                    "updated": self.artifact2.updated.isoformat().replace("+00:00", "Z"),
                },
            ],
        )

    def test_list_agent(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-artifacts", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    "content_type": "application/json",
                    "created": self.artifact1.created.isoformat().replace("+00:00", "Z"),
                    "id": str(self.artifact1.id),
                    "path": "path/to/file.json",
                    "s3_put_url": None,
                    "size": 42,
                    "updated": self.artifact1.updated.isoformat().replace("+00:00", "Z"),
                },
                {
                    "content_type": "text/plain",
                    "created": self.artifact2.created.isoformat().replace("+00:00", "Z"),
                    "id": str(self.artifact2.id),
                    "path": "some/text.txt",
                    "s3_put_url": None,
                    "size": 1337,
                    "updated": self.artifact2.updated.isoformat().replace("+00:00", "Z"),
                },
            ],
        )

    def test_list_other_task_no_access(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:task-artifacts", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f'Ponos {self.task3.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_other_task_with_access(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:task-artifacts", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f'Ponos {self.task2.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_task(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-artifacts", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f'Ponos {self.task1.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    "content_type": "application/json",
                    "created": self.artifact1.created.isoformat().replace("+00:00", "Z"),
                    "id": str(self.artifact1.id),
                    "path": "path/to/file.json",
                    "s3_put_url": None,
                    "size": 42,
                    "updated": self.artifact1.updated.isoformat().replace("+00:00", "Z"),
                },
                {
                    "content_type": "text/plain",
                    "created": self.artifact2.created.isoformat().replace("+00:00", "Z"),
                    "id": str(self.artifact2.id),
                    "path": "some/text.txt",
                    "s3_put_url": None,
                    "size": 1337,
                    "updated": self.artifact2.updated.isoformat().replace("+00:00", "Z"),
                },
            ],
        )

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={"path": "some/path.txt", "content_type": "text/plain", "size": 1000},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_forbids_admin(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={"path": "some/path.txt", "content_type": "text/plain", "size": 1000},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_forbids_user(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={"path": "some/path.txt", "content_type": "text/plain", "size": 1000},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_unassigned_agent(self):
        self.assertNotEqual(self.task1.agent, self.agent)

        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={"path": "some/path.txt", "content_type": "text/plain", "size": 1000},
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_assigned_agent(self):
        self.task1.agent = self.agent
        self.task1.save()

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={"path": "some/path.txt", "content_type": "text/plain", "size": 1000},
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # Check response has a valid S3 put URL, without matching the parameters in the querystring
        data = response.json()
        s3_put_url = data.get("s3_put_url")
        self.assertIsNotNone(s3_put_url)
        del data["s3_put_url"]
        self.assertTrue(
            s3_put_url.startswith(
                f"http://s3/ponos-artifacts/{self.task1.id}/some/path.txt"
            )
        )

        # An artifact has been created
        artifact = self.task1.artifacts.get(path='some/path.txt')

        self.assertDictEqual(
            response.json(),
            {
                "content_type": "text/plain",
                "created": artifact.created.isoformat().replace("+00:00", "Z"),
                "id": str(artifact.id),
                "path": "some/path.txt",
                "size": 1000,
                "updated": artifact.updated.isoformat().replace("+00:00", "Z"),
            },
        )

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_unique_path(self):
        self.task1.agent = self.agent
        self.task1.save()
        self.task1.artifacts.create(
            path="some/path.txt",
            content_type="something",
            size=42,
        )

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={
                    "path": "some/path.txt",
                    "content_type": "text/plain",
                    "size": 10243,
                },
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(
            response.json(), {"path": ["An artifact with this path already exists"]}
        )

    @override_settings()
    def test_create_size_limits(self):
        self.task1.agent = self.agent
        self.task1.save()

        params = [
            (None, -42, "Ensure this value is greater than or equal to 1."),
            (None, 0, "Ensure this value is greater than or equal to 1."),
            (
                None,
                5 * 1024**3 + 1,
                "Ensure this value is less than or equal to 5368709120.",
            ),
            (123456789000, -42, "Ensure this value is greater than or equal to 1."),
            (123456789000, 0, "Ensure this value is greater than or equal to 1."),
            (
                123456789000,
                987654321000,
                "Ensure this value is less than or equal to 123456789000.",
            ),
        ]

        for size_setting, size, expected_error in params:
            with self.subTest(size=size), self.assertNumQueries(3):
                settings.PONOS_ARTIFACT_MAX_SIZE = size_setting
                response = self.client.post(
                    reverse("api:task-artifacts", args=[self.task1.id]),
                    data={
                        "path": "some/path.txt",
                        "content_type": "text/plain",
                        "size": size,
                    },
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {"size": [expected_error]})

    def test_create_other_task(self):
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={
                    "path": "some/path.txt",
                    "content_type": "text/plain",
                    "size": 1000,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task2.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_task(self):
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={
                    "path": "some/path.txt",
                    "content_type": "text/plain",
                    "size": 1000,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        s3_put_url = data.pop("s3_put_url")
        self.assertTrue(
            s3_put_url.startswith(
                f"http://s3/ponos-artifacts/{self.task1.id}/some/path.txt"
            )
        )

        artifact = self.task1.artifacts.get(path='some/path.txt')
        self.assertDictEqual(
            data,
            {
                "content_type": "text/plain",
                "created": artifact.created.isoformat().replace("+00:00", "Z"),
                "id": str(artifact.id),
                "path": "some/path.txt",
                "size": 1000,
                "updated": artifact.updated.isoformat().replace("+00:00", "Z"),
            },
        )

    def test_download_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"])
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_requires_process_guest(self):
        self.corpus.memberships.filter(user=self.user).delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_process_level_corpus(self):
        self.client.force_login(self.user)

        for role in Role:
            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)

                with self.assertNumQueries(5):
                    response = self.client.get(
                        reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
                    )
                    self.assertEqual(response.status_code, status.HTTP_302_FOUND)

                self.assertTrue(response.has_header("Location"))
                self.assertTrue(
                    response["Location"].startswith(
                        f"http://s3/ponos-artifacts/{self.task1.id}/path/to/file.json"
                    )
                )

    def test_download_process_level_repo(self):
        self.client.force_login(self.user)
        membership = self.repo.memberships.create(user=self.user, level=Role.Guest.value)

        task = self.repository_process.tasks.create(run=0, depth=0, slug='a')
        task.artifacts.create(path='path/to/file.json', content_type='application/json', size=42)

        for role in Role:
            with self.subTest(role=role):
                membership.level = role.value
                membership.save()

                with self.assertNumQueries(6):
                    response = self.client.get(
                        reverse("api:task-artifact-download", args=[task.id, "path/to/file.json"]),
                    )
                    self.assertEqual(response.status_code, status.HTTP_302_FOUND)

                self.assertTrue(response.has_header("Location"))
                self.assertTrue(
                    response["Location"].startswith(
                        f"http://s3/ponos-artifacts/{task.id}/path/to/file.json"
                    )
                )

    def test_download_not_found(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "nope.xxx"]),
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_download_agent(self):
        # Being assigned to the task is not required
        self.assertNotEqual(self.task1.agent, self.agent)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        self.assertTrue(response.has_header("Location"))
        self.assertTrue(
            response["Location"].startswith(
                f"http://s3/ponos-artifacts/{self.task1.id}/path/to/file.json"
            )
        )

    def test_download_task(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
                HTTP_AUTHORIZATION=f'Ponos {self.task1.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        self.assertTrue(response.has_header("Location"))
        self.assertTrue(
            response["Location"].startswith(
                f"http://s3/ponos-artifacts/{self.task1.id}/path/to/file.json"
            )
        )

    def test_download_other_task_with_access(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
                HTTP_AUTHORIZATION=f'Ponos {self.task2.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertTrue(response.has_header("Location"))
        self.assertTrue(
            response["Location"].startswith(
                f"http://s3/ponos-artifacts/{self.task1.id}/path/to/file.json"
            )
        )

    def test_download_other_task_no_access(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]),
                HTTP_AUTHORIZATION=f'Ponos {self.task3.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
