import base64
import logging
import urllib.parse
import uuid
from abc import ABC, abstractmethod

from django.conf import settings
from django.urls import reverse
from gitlab import Gitlab, GitlabCreateError, GitlabGetError
from rest_framework.exceptions import APIException, AuthenticationFailed, NotAuthenticated, ValidationError

from arkindex.process.utils import get_default_farm

logger = logging.getLogger(__name__)


class GitProvider(ABC):

    display_name = None

    def __init__(self, credentials=None):
        if credentials:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
        self.credentials = credentials

    @abstractmethod
    def list_repos(self, query=None):
        """
        List all repositories or filter with a search query.
        """

    @abstractmethod
    def get_latest_commit_sha(self, repo):
        """
        Retrieve latest commit sha from the default branch on a given repository.
        """

    @abstractmethod
    def create_repo(self, **kwargs):
        """
        Create a Repository instance from an external repository
        """

    @abstractmethod
    def create_hook(self, repository, project_id=None, base_url=None):
        """Create a webhook to receive events from the project"""

    def update_or_create_ref(self, repo, revision, name, type):
        """
        Update or create a GitRef on a given repository:
        - If a GitRef with given name already exists on the given repository, we only update the
        linked revision with the given one
        - Else a GitRef is created on the given repository thanks to given name, type and revision
        """
        from arkindex.process.models import GitRef
        try:
            ref = repo.refs.get(name=name)
            ref.revision = revision
            ref.save()
        except GitRef.DoesNotExist:
            # Limit the name length to the max size of GitRef name field
            repo.refs.create(name=name[:250], type=type, revision=revision)

    def get_or_create_revision(self, repo, sha, save=True):
        from arkindex.process.models import Revision
        try:
            return self.get_revision(repo, sha), False
        except Revision.DoesNotExist:
            return self.create_revision(repo, sha, save=save), True

    def get_revision(self, repo, sha):
        return repo.revisions.get(hash=sha)

    @abstractmethod
    def create_revision(self, repo, sha, save=True):
        """
        Create a Revision instance for a given commit hash of a given repository.
        """

    @abstractmethod
    def handle_webhook(self, repo, request):
        """
        Handle a webhook event on a given repository.
        """

    @abstractmethod
    def to_dict(self, project):
        """
        Returns a dict specifying id, name and url from a repository
        """

    def start_imports(self, rev):
        """
        Create and start a process to build new worker(s) from a repository revision
        """
        from arkindex.process.models import ProcessMode
        mode = ProcessMode.Repository
        user = rev.repo.credentials.user
        refs = ", ".join(rev.refs.values_list('name', flat=True))
        if not refs:
            refs = rev.hash
        # Limit the name length to the max size of Process name field
        name = f"Import {refs} from {rev.repo.url}"[:100]

        farm = get_default_farm()
        if not farm.is_available(user):
            raise ValidationError("The owner of the OAuth credentials does not have access to the default farm.")

        process = rev.processes.create(
            creator=user,
            mode=mode,
            name=name,
            farm=farm,
        )
        process.run()
        return process


class GitLabProvider(GitProvider):

    display_name = "GitLab"

    def _get_gitlab_client(self, credentials):
        if credentials.expired:
            credentials.provider.refresh_token()

        return Gitlab(credentials.provider_url, oauth_token=credentials.token)

    def _try_get_project(self, gl, id):
        try:
            return gl.projects.get(id)
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

    def _get_project_from_repo(self, repo):
        assert repo.credentials, "Missing Gitlab credentials"

        gl = self._get_gitlab_client(repo.credentials)
        return self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))

    def list_repos(self, query=None):
        if not self.credentials:
            raise NotAuthenticated
        gl = self._get_gitlab_client(self.credentials)
        # Creating a webhook on a repo requires Maintainer (40) or Owner (50) access levels
        # See https://docs.gitlab.com/ce/api/members.html#valid-access-levels
        return gl.projects.list(min_access_level=40, search=query, get_all=False)

    def get_latest_commit_sha(self, repo):
        project = self._get_project_from_repo(repo)
        # Since ref_name isn't specified, we will use the repository default branch
        # See : https://docs.gitlab.com/ee/api/commits.html#list-repository-commits
        return project.commits.list(per_page=1, get_all=False)[0].id

    def to_dict(self, project):
        return {
            'id': project.id,
            'name': project.name_with_namespace,
            'url': project.web_url
        }

    def create_repo(self, id=None, **kwargs):
        if not self.credentials:
            raise NotAuthenticated()

        gl = self._get_gitlab_client(self.credentials)
        project = self._try_get_project(gl, int(id))

        from arkindex.process.models import Repository
        if Repository.objects.filter(url=project.web_url).exists():
            raise ValidationError("A repository with this URL already exists")

        # Determine the user's access level on this project
        # When it is inherited from the group and not overridden in the project, project_access is not defined.
        # When the project isn't in a group, or the user is added to a specific project in a group,
        # group_access is not defined. project_access overrides group_access.
        access_level = 0
        if project.permissions.get('group_access'):
            access_level = project.permissions['group_access']['access_level']
        if project.permissions.get('project_access'):
            access_level = project.permissions['project_access']['access_level']

        # Maintainer level (40) is required
        if access_level < 40:
            raise ValidationError("Maintainer or Owner access is required to add a GitLab repository")

        repo = self.credentials.repos.create(
            url=project.web_url,
            hook_token=base64.b64encode(uuid.uuid4().bytes).decode('utf-8'),
        )

        self.create_hook(repo, project_id=int(id))

        return repo

    def create_hook(self, repository, project_id=None, base_url=None):
        """
        Configure the Gitlab hook to get events for a project.

        If `project_id` is set, then the project is retrieved directly from its GitLab project ID
        instead of matched using its path.

        The webhook's URL will use `base_url` as its base URL if it is set, and otherwise
        falls back on settings.BACKEND_PUBLIC_URL_OAUTH first, then settings.PUBLIC_HOSTNAME.
        """
        # Load project using a project ID or its repo path
        gitlab = Gitlab(repository.credentials.provider_url, oauth_token=repository.credentials.token)
        if project_id:
            project = self._try_get_project(gitlab, project_id)
        else:
            path = urllib.parse.urlparse(repository.url).path[1:]
            project = self._try_get_project(gitlab, path)

        if base_url is None:
            base_url = settings.BACKEND_PUBLIC_URL_OAUTH or settings.PUBLIC_HOSTNAME

        if base_url is None:
            raise APIException('Either the `base_url` argument, settings.BACKEND_PUBLIC_URL_OAUTH or settings.PUBLIC_HOSTNAME must be set.')

        url = urllib.parse.urljoin(base_url, reverse('api:import-hook', kwargs={'pk': repository.id}))
        logger.info(f"Webhook will be created as {url}")

        # Delete already configured hooks to be able to update
        for hook in project.hooks.list(all=True):
            if hook.url == url:
                hook.delete()
                logger.info(f"Deleted existing hook {hook.id}")

        try:
            # Create a new hook
            hook = project.hooks.create({
                'url': url,
                'push_events': True,
                "tag_push_events": True,
                'token': repository.hook_token,
            })
            logger.info(f"Created new hook {hook.id}")
        except GitlabCreateError as e:
            raise APIException("Error while creating GitLab hook: {}".format(str(e)))

    def create_revision(self, repo, sha, save=True):
        from arkindex.process.models import GitRefType, Revision

        project = self._get_project_from_repo(repo)
        commit = project.commits.get(sha)

        rev = Revision(
            repo=repo,
            hash=sha,
            message=commit.message,
            author=commit.author_name,
        )
        if save:
            rev.save()
            for ref in commit.refs():
                try:
                    ref_type = GitRefType(ref['type'])
                    self.update_or_create_ref(repo, rev, ref['name'], ref_type)
                except ValueError:
                    logger.warning(f'Git reference with type {ref["type"]} was ignored during revision creation')
                    continue

        return rev

    def update_revision_references(self, repo, sha):
        """
        Update all references for a specific revision
        This is needed when a tag is pushed after its commit
        is already ingested
        """
        from arkindex.process.models import GitRefType

        project = self._get_project_from_repo(repo)
        commit = project.commits.get(sha)

        rev = repo.revisions.get(hash=sha)
        for ref in commit.refs():
            try:
                ref_type = GitRefType(ref['type'])
                self.update_or_create_ref(repo, rev, ref['name'], ref_type)
            except ValueError:
                logger.warning(f'Git reference with type {ref["type"]} was ignored during revision creation')
                continue

    def update_repository_references(self, repo):
        """
        List all available references (branches and tags) on a remote repository
        and create new references or update existing ones
        New import process are started for new references
        """
        from arkindex.process.models import GitRefType
        project = self._get_project_from_repo(repo)

        def _update_reference(ref, ref_type):
            assert isinstance(ref_type, GitRefType)
            try:
                rev, created = repo.revisions.get_or_create(
                    hash=ref.commit['id'],
                    defaults={
                        'message': ref.commit['message'],
                        'author': ref.commit['author_email'],
                    }
                )
                self.update_or_create_ref(repo, rev, ref.name, ref_type)
                if created:
                    logger.info(f"Starting import for {ref_type.value} {ref.name}")
                    self.start_imports(rev)
            except ValueError:
                logger.warning(f'{ref_type.value} {ref.name} was ignored during revision creation')

        for branch in project.branches.list(get_all=True):
            _update_reference(branch, GitRefType.Branch)

        for tag in project.tags.list(get_all=True):
            _update_reference(tag, GitRefType.Tag)

    def handle_webhook(self, repo, request):
        from arkindex.process.models import GitRefType

        if 'HTTP_X_GITLAB_EVENT' not in request.META:
            raise ValidationError("Missing GitLab event type")

        if request.META['HTTP_X_GITLAB_EVENT'] not in ('Push Hook', 'Tag Push Hook'):
            raise ValidationError("Unsupported GitLab event type")

        if 'HTTP_X_GITLAB_TOKEN' not in request.META:
            raise NotAuthenticated("Missing GitLab secret token")
        if request.META['HTTP_X_GITLAB_TOKEN'] != repo.hook_token:
            raise AuthenticationFailed("Invalid GitLab secret token")

        if not isinstance(request.data, dict) or 'object_kind' not in request.data:
            raise ValidationError('Bad payload format')

        kind = request.data.get('object_kind')
        if kind not in ('push', 'tag_push'):
            raise ValidationError("Unsupported GitLab event type")

        sha = request.data.get('checkout_sha')
        if kind == 'tag_push':
            if sha:
                # When a commit SHA is present, we need to add a reference on an existing commit
                self.update_revision_references(repo, sha)

            else:
                # If there isn't any SHA it means that a tag was deleted
                ref = request.data.get('ref')
                if not ref:
                    raise ValidationError('Missing tag reference')

                # Delete existing tag
                tag_name = ref[10:] if ref.startswith('refs/tags/') else ref
                repo.refs.filter(name=tag_name, type=GitRefType.Tag).delete()

            return

        if not sha:
            # If there isn't any SHA it means that a branch was deleted
            ref = request.data.get('ref')
            if not ref:
                raise ValidationError('Missing branch reference')

            # Delete existing branch
            branch_name = ref[11:] if ref.startswith('refs/heads/') else ref
            repo.refs.filter(name=branch_name, type=GitRefType.Branch).delete()

            return

        # Already took care of this event
        if repo.revisions.filter(hash=sha).exists():
            return

        rev = self.create_revision(repo, sha)

        self.start_imports(rev)


git_providers = [
    GitLabProvider,
]
oauth_to_git = {
    "gitlab": GitLabProvider,
}


def get_provider(name):
    return next(filter(lambda p: p.__name__ == name, git_providers), None)


def from_oauth(name):
    return oauth_to_git.get(name)
