import base64
import urllib
import uuid
from collections import defaultdict
from enum import Enum
from textwrap import dedent

from django.db import transaction
from django.db.models import Max, Q
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError

from arkindex.ponos.models import Task
from arkindex.ponos.utils import get_process_from_task_auth
from arkindex.process.models import (
    CorpusWorkerVersion,
    FeatureUsage,
    GitRef,
    Process,
    ProcessMode,
    Repository,
    Revision,
    Worker,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.serializers.git import GitRefSerializer, RevisionWithRefsSerializer
from arkindex.process.utils import hash_object
from arkindex.project.serializer_fields import EnumField
from arkindex.training.serializers import ModelVersionLightSerializer
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


class ExistingWorkerException(Exception):
    def __init__(self, worker):
        self.worker = worker


class WorkerSerializer(serializers.ModelSerializer):
    """
    Serialize a simple repository worker
    """
    # We want to only have to specify the type as a slug (char) in stead of the actual workerType id
    type = serializers.SlugField()
    repository_id = serializers.UUIDField(read_only=True, allow_null=True)

    class Meta:
        model = Worker
        fields = ('id', 'name', 'type', 'slug', 'repository_id')

    def create(self, validated_data):
        """
        Retrieve or create the worker and its type.
        In case of using Ponos auth, the worker is linked to the process repository
        """
        request = self.context['request']
        repository = None
        if isinstance(request.auth, Task):
            process = get_process_from_task_auth(request)
            if process is None or process.mode != ProcessMode.Repository or process.revision_id is None:
                raise PermissionDenied(
                    'Workers can only be created on processes of type Repository with a'
                    ' revision while using the Ponos authentication.'
                )
            repository = process.revision.repo

            access_right = get_max_level(request.user, repository)
            if not access_right:
                raise NotFound
            elif access_right < Role.Admin.value:
                raise PermissionDenied(detail='You do not have admin access to the repository of this process.')

        with transaction.atomic(using='default'):
            # First ensure no worker exists with the given slug and repository
            existing_worker = (
                Worker.objects
                .using('default')
                .filter(repository=repository, slug=validated_data['slug'])
                .first()
            )
            if existing_worker:
                if not existing_worker.is_executable(user=request.user):
                    raise ValidationError({'slug': ['You cannot use this value.']})
                raise ExistingWorkerException(existing_worker)

            # Get or create the worker and its type
            worker_type, _ = WorkerType.objects.using('default').get_or_create(
                slug=validated_data['type'],
                defaults={
                    'display_name': validated_data['type'].capitalize()
                }
            )
            worker = Worker.objects.create(
                repository=repository,
                slug=validated_data['slug'],
                name=validated_data['name'],
                type=worker_type,
            )
            if repository is None:
                # Automatically grant an admin access to the creator of a local worker
                worker.memberships.create(user=request.user, level=Role.Admin.value)
            return worker


class WorkerLightSerializer(WorkerSerializer):
    """
    Serialize a worker without its repository ID
    """
    class Meta:
        model = Worker
        fields = ('id', 'name', 'type', 'slug')


class WorkerTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = WorkerType
        fields = (
            'id',
            'slug',
            'display_name'
        )


class UserConfigurationFieldType(Enum):
    Int = 'int'
    Float = 'float'
    String = 'string'
    Enum = 'enum'
    Boolean = 'bool'
    Dict = 'dict'
    List = 'list'


class UserConfigurationFieldSerializer(serializers.Serializer):
    title = serializers.CharField()
    type = EnumField(UserConfigurationFieldType)
    subtype = EnumField(UserConfigurationFieldType, required=False)
    required = serializers.BooleanField(default=False)
    choices = serializers.ListField(required=False, allow_empty=False, allow_null=True)

    def to_internal_value(self, data):
        errors = defaultdict(list)

        allowed_fields = ['title', 'type', 'required', 'default', 'choices', 'subtype']
        data_types = {
            UserConfigurationFieldType.Int: [serializers.IntegerField(), int],
            UserConfigurationFieldType.Float: [serializers.FloatField(), float],
            UserConfigurationFieldType.String: [serializers.CharField(), str],
            UserConfigurationFieldType.Enum: [serializers.ChoiceField(choices=[]), None],
            UserConfigurationFieldType.Boolean: [serializers.BooleanField(), bool],
            UserConfigurationFieldType.Dict: [serializers.DictField(child=serializers.CharField()), dict],
            UserConfigurationFieldType.List: [serializers.ListField(), list]
        }

        if not isinstance(data, dict):
            errors['__all__'] = [f'User configuration field definitions should be of type dict, not {type(data).__name__}.']
            raise ValidationError(errors)

        for field in data:
            if field not in allowed_fields:
                errors[field].append(
                    'Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices.'
                )

        default_value = data.get('default')
        data = super().to_internal_value(data)

        field_type = data.get('type')
        choices = data.get('choices')
        subtype = data.get('subtype')

        if field_type == UserConfigurationFieldType.List and not subtype:
            errors['subtype'].append('The "subtype" field must be set for "list" type properties.')
        # Handle subtypes
        if subtype is not None:
            if field_type != UserConfigurationFieldType.List:
                errors['subtype'].append('The "subtype" field can only be set for a "list" type property.')
            if subtype not in [UserConfigurationFieldType.Int, UserConfigurationFieldType.Float, UserConfigurationFieldType.String, UserConfigurationFieldType.Boolean]:
                errors['subtype'].append('Subtype can only be int, float, bool or string.')
        # Handle enums
        if choices is not None:
            if field_type != UserConfigurationFieldType.Enum:
                errors['choices'].append('The "choices" field can only be set for an "enum" type property.')
            # If the configuration parameter is of enum type, an eventual default value won't match the field type
            if default_value and default_value not in choices:
                errors['default'].append(f'{default_value} is not an available choice.')
        # Handle everything else
        if default_value is not None and field_type != UserConfigurationFieldType.Enum:
            try:
                data_type, data_class = data_types[field_type]
                if not isinstance(default_value, data_class):
                    raise ValidationError
                data_type.to_internal_value(default_value)
                # For lists, check that list elements are of given subtype
                if field_type == UserConfigurationFieldType.List and not errors.get('subtype'):
                    _, data_subclass = data_types[subtype]
                    if any(not isinstance(item, data_subclass) for item in default_value):
                        errors['default'].append(f'All items in the default value must be of type {data_subclass.__name__}.')
            except ValidationError:
                errors['default'].append(f'Default value is not of type {field_type.value}.')
            except KeyError:
                errors['default'].append(f'Cannot check type: {field_type.value}.')

        if errors:
            raise ValidationError(errors)
        return data


class WorkerVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a worker version
    """
    # State defaults to created when instantiating a WorkerVersion
    state = EnumField(WorkerVersionState, required=False)
    worker = WorkerLightSerializer(read_only=True)
    revision = RevisionWithRefsSerializer(required=False, read_only=True, allow_null=True)
    gpu_usage = EnumField(FeatureUsage, required=False, default=FeatureUsage.Disabled)
    model_usage = EnumField(FeatureUsage, required=False, default=FeatureUsage.Disabled)
    # Ensure worker version configuration JSON body is an object
    configuration = serializers.DictField()

    # Serialize worker with its basic information
    class Meta:
        model = WorkerVersion
        fields = (
            'id',
            'configuration',
            'revision',
            'version',
            'docker_image',
            'docker_image_iid',
            'docker_image_name',
            'state',
            'gpu_usage',
            'model_usage',
            'worker',
            'created',
        )
        read_only_fields = ('docker_image_name', 'revision', 'version')
        # Avoid loading all revisions and all Ponos artifacts when opening this endpoint in a browser
        extra_kwargs = {
            'revision': {'style': {'base_template': 'input.html'}},
            'docker_image': {'style': {'base_template': 'input.html'}},
        }

    def validate_configuration(self, configuration):
        errors = defaultdict(list)
        user_configuration = configuration.get('user_configuration')
        if not user_configuration:
            return configuration
        field = serializers.DictField(child=UserConfigurationFieldSerializer())
        try:
            field.to_internal_value(user_configuration)
        except ValidationError as e:
            errors['user_configuration'].append(e.detail)
        if errors:
            raise ValidationError(errors)
        return configuration

    def validate(self, data):
        # Assert that a version set to available either has a FK to a docker image artifact or an image name
        state = data.get('state') or self.instance and self.instance.state
        if not state:
            return data

        image = (
            data['docker_image'] if 'docker_image' in data
            else self.instance and self.instance.docker_image_id
        )
        image_iid = (
            data['docker_image_iid'] if 'docker_image_iid' in data
            else self.instance and self.instance.docker_image_iid
        )
        if state == WorkerVersionState.Available and not image and not image_iid:
            raise ValidationError({'state': [
                'A worker version must either have a Docker image artifact '
                'or a `docker_image_iid` set when updated to available.'
            ]})

        return data


class WorkerVersionCreateSerializer(WorkerVersionSerializer):
    revision_id = serializers.PrimaryKeyRelatedField(
        required=False,
        write_only=True,
        queryset=Revision.objects.all(),
        allow_null=True,
        style={'base_template': 'input.html'},
        help_text=dedent("""
            Git revision for this version.

            This field is required on workers linked to a repository.
            On other workers, it cannot be set as an automatic version number is attributed.
        """).strip(),
        source='revision'
    )

    class Meta (WorkerVersionSerializer.Meta):
        fields = WorkerVersionSerializer.Meta.fields + (
            'revision_id',
        )

    def validate_revision_id(self, revision):
        worker = self.context['worker']
        if worker.versions.using('default').filter(revision=revision).exists():
            raise ValidationError('A version of this worker already exists with this revision')
        if isinstance(self.context['request'].auth, Task):
            if worker.repository_id is None:
                # Task authentication is forbidden on workers not linked to a repository
                raise ValidationError(
                    'Task authentication requires to create a version on workers linked to a repository.'
                )
        elif worker.repository_id is not None:
            raise ValidationError(
                'Ponos authentication is required to create a version on a worker linked to a repository.'
            )
        if worker.repository is None:
            raise ValidationError('A revision cannot be set on a worker that is not linked to a repository.')
        if worker.repository_id != revision.repo_id:
            raise ValidationError('The revision must be part of the same repository as the worker.')
        return revision

    def validate(self, data):
        super().validate(data)
        worker = self.context['worker']
        revision = data.get('revision')
        if revision is None and self.context['worker'].repository_id is not None:
            raise ValidationError(
                {'revision_id': ['A revision must be set on a worker that is linked to a repository.']}
            )
        # Define a version number on workers that are not linked to a repository
        version = None
        if revision is None:
            last_version = worker.versions.using('default').aggregate(last_version=Max('version'))["last_version"]
            version = last_version + 1 if last_version else 1
        data['version'] = version
        data['worker_id'] = worker.id
        return data


class RepositorySerializer(serializers.ModelSerializer):
    """
    Serialize a repository
    """
    enabled = serializers.BooleanField(read_only=True)
    git_clone_url = serializers.SerializerMethodField()
    workers = WorkerSerializer(many=True, read_only=True)
    authorized_users = serializers.SerializerMethodField(read_only=True)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_git_clone_url(self, repository):
        # This check avoid to set git_clone_url when this serializer is used to list multiple
        # repositories because self.instance value would be a list, even with Ponos task authentication.
        # We also restrict to enabled repositories as disabled repos do not have OAuth credentials,
        # and restrict to the repository set on the task's process, as it should only clone that one.
        process = get_process_from_task_auth(self.context['request'])
        if (
            process
            and isinstance(self.instance, Repository)
            and self.instance.enabled
            and process.revision_id is not None
            and process.revision.repo_id == self.instance.id
        ):
            url = urllib.parse.urlparse(self.instance.url)
            return f"https://oauth2:{repository.credentials.token}@{url.netloc}{url.path}"
        return None

    def get_authorized_users(self, repo) -> int:
        count = getattr(repo, 'authorized_users', None)
        if count is None:
            count = repo.memberships.count()
        return count

    class Meta:
        model = Repository
        fields = (
            'id',
            'url',
            'enabled',
            'git_clone_url',
            'workers',
            'authorized_users',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'url': {'read_only': True}
        }


class WorkerActivitySerializer(serializers.ModelSerializer):
    """
    Serialize a worker activity on an element
    """
    state = EnumField(WorkerActivityState)
    element_id = serializers.UUIDField()
    process_id = serializers.PrimaryKeyRelatedField(
        queryset=Process.objects.all(),
        # Avoid loading up to 1000 processes when opening this endpoint in a browser
        style={'base_template': 'input.html'},
        required=False,
        help_text=dedent("""
            ID of the process on which the worker ran.

            Mandatory when updating an activity using WorkerVersion IDs, ignored with WorkerRun IDs.
        """),
    )
    configuration_id = serializers.UUIDField(read_only=True, allow_null=True)
    model_version_id = serializers.UUIDField(read_only=True, allow_null=True)
    worker_version_id = serializers.UUIDField(read_only=True)

    class Meta:
        model = WorkerActivity
        fields = (
            'element_id',
            'process_id',
            'configuration_id',
            'worker_version_id',
            'model_version_id',
            'created',
            'updated',
            'started',
            'state',
        )
        read_only_fields = (
            'created',
            'updated',
            'started',
            'configuration_id',
            'worker_version_id',
            'model_version_id',
        )

    def validate_process_id(self, process):
        request = self.context.get('request')
        if request and isinstance(request.auth, Task) and process.id != request.auth.process_id:
            raise serializers.ValidationError('Only WorkerActivities for the process of the currently authenticated task can be updated.')
        return process


class WorkerConfigurationListSerializer(serializers.ModelSerializer):
    configuration = serializers.DictField(allow_empty=False)

    class Meta:
        model = WorkerConfiguration
        fields = (
            'id',
            'name',
            'configuration',
            'archived'
        )

    def validate(self, data):
        # Check unique constraints
        name = data.get('name')
        configuration = data.get('configuration')
        configuration_hash = hash_object(configuration)

        # No further validation required if we only sent `archived`
        if not name and not configuration:
            return data

        if 'worker' in self.context:
            # CreateWorkerConfiguration provides the worker through the context
            worker = self.context['worker']
            data['worker_id'] = worker.id
        elif self.instance is not None:
            # Use the already set worker in (Partial)UpdateConfiguration to revalidate
            worker = self.instance.worker

        existing_configurations = list(worker.configurations.filter(Q(configuration_hash=configuration_hash) | Q(name=name)))
        errors = defaultdict(list)
        for config in existing_configurations:
            if configuration_hash == config.configuration_hash:
                errors['configuration'].append("A worker configuration with this configuration already exists for this worker.")
                errors['id'] = config.id
            if name == config.name:
                errors['name'].append("A worker configuration with this name already exists for this worker.")

        # Due to the defaultdict, writing e.g. len(errors['configuration']) > 1 would add {'configuration': []} to
        # the errors dict if the key is not already in the dict.
        if len(errors.get('name', [])) > 1 or len(errors.get('configuration', [])) > 1:
            # If there is more than one configuration with the same name and/or configuration as the one the user
            # is trying to create, raise a MultipleObjectsReturned exception because this shouldn't be possible
            # due to the unicity constraints. There can however be more than one worker configurations in
            # existing_configurations, if both a worker configuration with the same name and the same configuration
            # exist; but checking the length of existing_configurations wouldn't differentiate between this case
            # and an actual duplicated worker configuration.
            raise WorkerConfiguration.MultipleObjectsReturned

        # Archived configurations cannot be renamed, but un-archiving a configuration and renaming it at the same time is
        # possible. It's also possible to archive and rename a configuration simultaneously.
        if self.instance is not None and self.instance.archived and data.get('archived') is not False and name:
            errors['name'].append('Archived configurations cannot be renamed.')

        if errors:
            raise ValidationError(errors)

        return data


class WorkerConfigurationSerializer(WorkerConfigurationListSerializer):
    configuration = serializers.DictField(allow_empty=False, read_only=True)

    class Meta(WorkerConfigurationListSerializer.Meta):
        # The configuration cannot be updated
        read_only_fields = ('id', 'configuration')


class CorpusWorkerVersionSerializer(serializers.ModelSerializer):
    worker_version = WorkerVersionSerializer()
    model_version = ModelVersionLightSerializer(read_only=True, allow_null=True)
    worker_configuration = WorkerConfigurationSerializer(read_only=True, allow_null=True)

    class Meta:
        model = CorpusWorkerVersion
        fields = ('id', 'worker_version', 'model_version', 'worker_configuration')


class WorkerConfigurationExistsErrorSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing worker configuration, if the error comes from a duplicate configuration.")
    configuration = serializers.ListField(child=serializers.CharField(), required=False, help_text="Configuration error message.")
    name = serializers.ListField(child=serializers.CharField(), required=False, help_text="Name error message.")


class WorkerStatisticsSerializer(serializers.Serializer):
    """
    Serialize activity statistics of a worker version
    """
    worker_version_id = serializers.UUIDField(read_only=True)
    configuration_id = serializers.UUIDField(read_only=True, allow_null=True)
    model_version_id = serializers.UUIDField(read_only=True, allow_null=True)
    queued = serializers.IntegerField(read_only=True)
    started = serializers.IntegerField(read_only=True)
    processed = serializers.IntegerField(read_only=True)
    error = serializers.IntegerField(read_only=True)
    average_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        # Also document the timestamp format, specified by a random Django util function used by DRF:
        # https://github.com/django/django/blob/eccda63a49b92ab14d523d69da2a3b3246cf91d1/django/utils/duration.py#L18
        help_text=dedent('''
        Average time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        '''),
    )
    min_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent('''
        Shortest time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        '''),
    )
    max_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent('''
        Longest time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        '''),
    )
    median_processed_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent('''
        Median time spent processing each element, only for `Processed` worker activities.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        '''),
    )
    completion_estimate_time = serializers.DurationField(
        read_only=True,
        allow_null=True,
        help_text=dedent('''
        Estimated time for remaining elements to be processed based on median process time.

        Format: `[DD ]HH:MM:ss[.uuuuuu]`
        '''),
    )


class DockerWorkerVersionSerializer(serializers.ModelSerializer):

    repository_url = serializers.CharField(
        max_length=100,
        help_text="URL of the worker's repository. If the repository does not exist, it will be created."
    )
    worker_slug = serializers.CharField(
        max_length=100,
        help_text=dedent('''
        The slug/name of the worker to which a new version will be published.
        If such a worker does not exist, it will be created.
        '''),
    )
    revision_hash = serializers.CharField(max_length=50)
    revision_message = serializers.CharField(required=False, default="created from docker image")
    revision_author = serializers.CharField(max_length=50, required=False, default="default")
    revision_references = GitRefSerializer(
        many=True,
        required=False,
        help_text="Optional list of types and names to create GitRefs",
    )
    gpu_usage = EnumField(FeatureUsage, required=False, default=FeatureUsage.Disabled)
    model_usage = EnumField(FeatureUsage, required=False, default=FeatureUsage.Disabled)

    class Meta:
        model = WorkerVersion
        fields = (
            'configuration',
            'docker_image_iid',
            'gpu_usage',
            'model_usage',
            # Related fields
            'repository_url',
            'worker_slug',
            'revision_hash',
            'revision_message',
            'revision_author',
            'revision_references',
        )
        extra_kwargs = {
            # The Docker image tag is required in this endpoint
            'docker_image_iid': {
                'required': True,
                'allow_null': False,
                'allow_blank': False,
            }
        }

    @transaction.atomic
    def create(self, validated_data):
        """
        Retrieve or create the entire git stack in order to build a
        worker version corresponding to a given revision on the repository.
        """
        # Retrieve or create the Git repository
        repository, created_repo = Repository.objects.using('default').get_or_create(
            url=validated_data['repository_url'],
            defaults={
                # Generate a default hook token (DB constraint) even if no webhook is created
                'hook_token': base64.b64encode(uuid.uuid4().bytes).decode('utf-8')
            },
        )
        # Grant an admin access to the repository in case it got created
        if created_repo:
            repository.memberships.create(user=self.context['request'].user, level=Role.Admin.value)

        # Retrieve or create the revision (multiple workers may reference a same revision)
        revision, created = (
            Revision
            .objects
            .using('default')
            # The prefetch will only occur for an existing revision
            .prefetch_related('refs')
            .get_or_create(
                repo=repository,
                hash=validated_data['revision_hash'],
                defaults={
                    'message': validated_data['revision_message'],
                    'author': validated_data['revision_author'],
                },
            )
        )

        if created:
            existing_git_refs = set()
        else:
            existing_git_refs = {(ref.type, ref.name) for ref in revision.refs.all()}

        # Add Git references to the created revision
        new_git_refs = {
            (ref['type'], ref['name'])
            for ref in validated_data.get('revision_references', [])
        } - existing_git_refs

        GitRef.objects.bulk_create(
            (
                GitRef(revision=revision, repository=repository, type=ref_type, name=ref_name)
                for ref_type, ref_name in new_git_refs
            ),
            # When a GitRef already exists on the repo, update its revision to the current one.
            # If the revision already exists, this could mean that an earlier revision
            # could get a ref that it shouldn't have, for example when retrying a CI pipeline.
            update_conflicts=True,
            update_fields=['revision'],
            # With PostgreSQL, Django also requires specifying the fields
            # of the unique constraint that will cause a conflict.
            # Note that GitRefs are unique per name and repository, but not by type.
            # This means that a branch and a tag cannot have the same name,
            # or one of the two will be skipped without any error.
            unique_fields=['repository', 'name'],
        )

        # Use a specific worker type in case a worker must be created
        worker_type, _ = WorkerType.objects.using('default').get_or_create(slug='docker', defaults={'display_name': 'Docker'})

        # Retrieve or create the worker
        worker, _ = repository.workers.using('default').get_or_create(
            slug=validated_data['worker_slug'],
            repository=repository,
            defaults={
                'name': validated_data['worker_slug'],
                'type': worker_type,
            },
        )

        # Finally, create the worker version and mark it as available
        # If we are about the return an existing worker version, fetch the required data for the response
        version, created = (
            WorkerVersion
            .objects
            .using('default')
            .select_related('worker__type', 'worker__repository', 'revision__repo')
            .prefetch_related('revision__refs')
            .get_or_create(
                revision=revision,
                worker=worker,
                defaults={
                    'configuration': validated_data['configuration'],
                    'docker_image_iid': validated_data['docker_image_iid'],
                    'gpu_usage': validated_data['gpu_usage'],
                    'model_usage': validated_data['model_usage'],
                    'state': WorkerVersionState.Available,
                },
            )
        )

        # Abort in case a version already exists with that revision and worker,
        # unless this API call has caused some new GitRefs to be added on the existing revision.
        if not created and not new_git_refs:
            raise ValidationError({
                '__all__': ['A worker version already exists on this revision and this worker slug.']
            })

        return version
