from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import FeatureUsage, GitRefType, ProcessMode, Repository, Worker, WorkerType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, Scope


class TestDockerWorkerVersion(FixtureAPITestCase):
    """
    Test endpoint to create a worker version from a docker image
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(url='http://my_repo.fake/workers/worker')
        cls.rev = cls.repo.revisions.get()
        cls.worker = Worker.objects.get(slug='reco')
        cls.version = cls.worker.versions.get()
        cls.user.user_scopes.create(scope=Scope.CreateDockerWorkerVersion)
        cls.worker_type = WorkerType.objects.create(slug='docker', display_name='Docker')

    def test_create_requires_login(self):
        response = self.client.post(reverse('api:version-from-docker'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_create_deny_ponos_auth(self):
        """
        Ponos task authentication is disabled on that endpoint
        """
        process = self.rev.processes.create(
            mode=ProcessMode.Repository,
            creator=self.user,
            farm=Farm.objects.first(),
        )
        process.run()
        task = process.tasks.get()
        response = self.client.post(
            reverse('api:version-from-docker'),
            HTTP_AUTHORIZATION=f'Ponos {task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:version-from-docker'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_create_requires_scope(self):
        """
        A specific user scope is required to create a worker version from a docker image
        """
        self.user.user_scopes.get(scope=Scope.CreateDockerWorkerVersion).delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:version-from-docker'))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {
            'detail': 'You do not have permission to perform this action.',
            'missing_scopes': [Scope.CreateDockerWorkerVersion.value],
        })

    def test_create_admin_no_scope(self):
        """
        User scope is not required for instance admins
        """
        self.assertEqual(self.superuser.user_scopes.count(), 0)
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('api:version-from-docker'),
            data={
                'configuration': {'test': 'A'},
                'docker_image_iid': 'a_docker_image',
                'repository_url': self.repo.url,
                'revision_hash': 'new_revision_hash',
                'worker_slug': self.worker.slug,
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:version-from-docker'))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                'configuration': ['This field is required.'],
                'docker_image_iid': ['This field is required.'],
                'repository_url': ['This field is required.'],
                'revision_hash': ['This field is required.'],
                'worker_slug': ['This field is required.'],
            },
        )

    def test_create_fields_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': 'str',
                    'docker_image_iid': [],
                    'repository_url': [],
                    'revision_hash': '',
                    'worker_slug': [],
                    'revision_message': [],
                    'revision_author': [],
                    'revision_references': [{'a': 133}],
                    'gpu_usage': 'AAAAAAAA',
                    'model_usage': [],
                    # Only revision can be set using this endpoint
                    'version': 42,
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'docker_image_iid': ['Not a valid string.'],
            'gpu_usage': ['Value is not of type FeatureUsage'],
            'model_usage': ['Value is not of type FeatureUsage'],
            'repository_url': ['Not a valid string.'],
            'revision_author': ['Not a valid string.'],
            'revision_hash': ['This field may not be blank.'],
            'revision_message': ['Not a valid string.'],
            'revision_references': [{
                'name': ['This field is required.'],
                'type': ['This field is required.']
            }],
            'worker_slug': ['Not a valid string.'],
        })

    def test_create_duplicated(self):
        """
        No worker version can be created with an existing revision hash and worker slug
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': {},
                    'docker_image_iid': 'some_docker_image',
                    'repository_url': self.repo.url,
                    'revision_hash': self.version.revision.hash,
                    'worker_slug': self.version.worker.slug,
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['A worker version already exists on this revision and this worker slug.']
        })

    def test_create_default_message(self):
        """
        A new version can be published with only the required fields
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': {'test': 'A'},
                    'docker_image_iid': 'a_docker_image',
                    'repository_url': self.repo.url,
                    'revision_hash': 'new_revision_hash',
                    'worker_slug': self.worker.slug,
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_revision = self.repo.revisions.get(hash='new_revision_hash')
        new_version = new_revision.versions.get()
        self.assertEqual(new_version.worker, self.worker)
        self.assertDictEqual(response.json(), {
            'id': str(new_version.id),
            'configuration': {'test': 'A'},
            'docker_image': None,
            'docker_image_iid': 'a_docker_image',
            'docker_image_name': new_version.docker_image_name,
            'gpu_usage': FeatureUsage.Disabled.value,
            'model_usage': FeatureUsage.Disabled.value,
            'revision': {
                'id': str(new_revision.id),
                'author': 'default',
                'commit_url': 'http://my_repo.fake/workers/worker/commit/new_revision_hash',
                'created': new_revision.created.isoformat().replace('+00:00', 'Z'),
                'hash': 'new_revision_hash',
                'message': 'created from docker image',
                'refs': []
            },
            'version': None,
            'created': new_version.created.isoformat().replace('+00:00', 'Z'),
            'state': 'available',
            'worker': {
                'id': str(self.worker.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer'
            }
        })

    def test_create(self):
        """
        A new version can be published with the optional values
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': {'test': 'A'},
                    'docker_image_iid': 'e' * 512,
                    'repository_url': self.repo.url,
                    'revision_hash': 'new_revision_hash',
                    'worker_slug': self.worker.slug,
                    'revision_message': 'Bruce was very clever',
                    'revision_author': 'Iwan Roberts',
                    'revision_references': [
                        {'type': 'branch', 'name': 'master'},
                        {'type': 'tag', 'name': '2.0'},
                    ],
                    'gpu_usage': FeatureUsage.Required.value,
                    'model_usage': FeatureUsage.Supported.value,
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_revision = self.repo.revisions.get(hash='new_revision_hash')
        refs = list(new_revision.refs.all())
        self.assertDictEqual(
            {ref.type: ref.name for ref in refs},
            {GitRefType.Branch: 'master', GitRefType.Tag: '2.0'},
        )
        new_version = new_revision.versions.get()
        self.assertDictEqual(response.json(), {
            'id': str(new_version.id),
            'configuration': {'test': 'A'},
            'docker_image': None,
            'docker_image_iid': 'e' * 512,
            'docker_image_name': new_version.docker_image_name,
            'gpu_usage': FeatureUsage.Required.value,
            'model_usage': FeatureUsage.Supported.value,
            'revision': {
                'id': str(new_revision.id),
                'author': 'Iwan Roberts',
                'commit_url': 'http://my_repo.fake/workers/worker/commit/new_revision_hash',
                'created': new_revision.created.isoformat().replace('+00:00', 'Z'),
                'hash': 'new_revision_hash',
                'message': 'Bruce was very clever',
                'refs': [
                    {'id': str(ref.id), 'name': ref.name, 'type': ref.type.value}
                    for ref in refs
                ],
            },
            'version': None,
            'created': new_version.created.isoformat().replace('+00:00', 'Z'),
            'state': 'available',
            'worker': {
                'id': str(self.worker.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer'
            }
        })
        # Existing repository memberships are not updated
        self.assertFalse(self.repo.memberships.exists())

    def test_create_update_git_ref(self):
        """
        Existing GitRefs on the repository on a different revision
        are updated to the new revision
        """
        self.assertFalse(self.repo.refs.exists())
        branch_ref = self.repo.refs.create(
            revision=self.rev,
            type=GitRefType.Branch,
            name='master',
        )
        tag_ref = self.repo.refs.create(
            revision=self.rev,
            type=GitRefType.Tag,
            name='2.0',
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': {'test': 'A'},
                    'docker_image_iid': 'a_docker_image',
                    'repository_url': self.repo.url,
                    'revision_hash': 'new_revision_hash',
                    'worker_slug': self.worker.slug,
                    'revision_message': 'Bruce was very clever',
                    'revision_author': 'Iwan Roberts',
                    'revision_references': [
                        {'type': 'branch', 'name': 'master'},
                        {'type': 'tag', 'name': '2.0'},
                    ],
                    'gpu_usage': FeatureUsage.Required.value,
                    'model_usage': FeatureUsage.Required.value,
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_revision = self.repo.revisions.get(hash='new_revision_hash')

        branch_ref.refresh_from_db()
        tag_ref.refresh_from_db()
        self.assertEqual(branch_ref.revision, new_revision)
        self.assertEqual(tag_ref.revision, new_revision)

        # No new refs were created
        self.assertEqual(self.repo.refs.count(), 2)

        data = response.json()
        # GitRefs are not sorted in the API output, so we have to assert separately
        refs_data = data['revision'].pop('refs')
        self.assertCountEqual(refs_data, [
            {
                'id': str(tag_ref.id),
                'name': '2.0',
                'type': 'tag',
            },
            {
                'id': str(branch_ref.id),
                'name': 'master',
                'type': 'branch',
            },
        ])

        new_version = new_revision.versions.get()
        self.assertDictEqual(response.json(), {
            'id': str(new_version.id),
            'configuration': {'test': 'A'},
            'docker_image': None,
            'docker_image_iid': 'a_docker_image',
            'docker_image_name': new_version.docker_image_name,
            'gpu_usage': FeatureUsage.Required.value,
            'model_usage': FeatureUsage.Required.value,
            'revision': {
                'id': str(new_revision.id),
                'author': 'Iwan Roberts',
                'commit_url': 'http://my_repo.fake/workers/worker/commit/new_revision_hash',
                'created': new_revision.created.isoformat().replace('+00:00', 'Z'),
                'hash': 'new_revision_hash',
                'message': 'Bruce was very clever',
            },
            'version': None,
            'created': new_version.created.isoformat().replace('+00:00', 'Z'),
            'state': 'available',
            'worker': {
                'id': str(self.worker.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer'
            }
        })
        # Existing repository memberships are not updated
        self.assertFalse(self.repo.memberships.exists())

    def test_create_add_git_ref(self):
        """
        If a worker version already exists, the call still succeeds
        when it creates a new GitRef
        """
        self.client.force_login(self.user)
        self.assertFalse(self.version.revision.refs.exists())
        branch_ref = self.version.revision.refs.create(
            name='fix-typo',
            type=GitRefType.Branch,
            repository=self.version.revision.repo,
        )

        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': {},
                    'docker_image_iid': 'some_docker_image',
                    'repository_url': self.repo.url,
                    'revision_hash': self.version.revision.hash,
                    'worker_slug': self.version.worker.slug,
                    'revision_references': [
                        {'type': 'tag', 'name': '9.9.9.9.9.9'},
                    ],
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        tag_ref = self.version.revision.refs.get(type=GitRefType.Tag)

        self.assertDictEqual(response.json(), {
            'id': str(self.version.id),
            'configuration': {'test': 42},
            'docker_image': str(self.version.docker_image_id),
            'docker_image_iid': None,
            'docker_image_name': self.version.docker_image_name,
            'gpu_usage': FeatureUsage.Disabled.value,
            'model_usage': FeatureUsage.Disabled.value,
            'revision': {
                'id': str(self.version.revision.id),
                'author': 'Test user',
                'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                'created': self.version.revision.created.isoformat().replace('+00:00', 'Z'),
                'hash': '1337',
                'message': 'My w0rk3r',
                'refs': [
                    {
                        'id': str(branch_ref.id),
                        'name': 'fix-typo',
                        'type': 'branch',
                    },
                    {
                        'id': str(tag_ref.id),
                        'name': '9.9.9.9.9.9',
                        'type': 'tag',
                    },
                ],
            },
            'version': None,
            'created': self.version.created.isoformat().replace('+00:00', 'Z'),
            'state': 'available',
            'worker': {
                'id': str(self.worker.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer'
            }
        })

    def test_create_git_stack(self):
        """
        All git references can be created including repository, revision, gitrefs and worker
        including the default worker type
        """
        self.worker_type.delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(29):
            response = self.client.post(
                reverse('api:version-from-docker'),
                data={
                    'configuration': {'key': 'value'},
                    'docker_image_iid': 'docker_image_42',
                    'repository_url': 'https://gitlab.test.arkindex.org/project/',
                    'revision_hash': 'deadbeef',
                    'worker_slug': 'new_gen_classifier',
                    'revision_references': [
                        {'type': 'branch', 'name': 'master'},
                    ],
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_repo = Repository.objects.get(url='https://gitlab.test.arkindex.org/project/')
        new_revision = new_repo.revisions.get(hash='deadbeef')
        new_ref = new_revision.refs.get()
        new_version = new_revision.versions.get()
        new_worker_type = new_version.worker.type
        self.assertDictEqual(response.json(), {
            'id': str(new_version.id),
            'configuration': {'key': 'value'},
            'docker_image': None,
            'docker_image_iid': 'docker_image_42',
            'docker_image_name': new_version.docker_image_name,
            'gpu_usage': FeatureUsage.Disabled.value,
            'model_usage': FeatureUsage.Disabled.value,
            'revision': {
                'id': str(new_revision.id),
                'author': 'default',
                'commit_url': 'https://gitlab.test.arkindex.org/project/commit/deadbeef',
                'created': new_revision.created.isoformat().replace('+00:00', 'Z'),
                'hash': 'deadbeef',
                'message': 'created from docker image',
                'refs': [{'id': str(new_ref.id), 'name': new_ref.name, 'type': new_ref.type.value}],
            },
            'state': 'available',
            'version': None,
            'created': new_version.created.isoformat().replace('+00:00', 'Z'),
            'worker': {
                'id': str(new_version.worker.id),
                'name': 'new_gen_classifier',
                'slug': 'new_gen_classifier',
                'type': str(new_worker_type),
            }
        })
        # Check attributes directly from the DB
        self.assertEqual(new_repo.url, 'https://gitlab.test.arkindex.org/project/')
        self.assertEqual(new_revision.author, 'default')
        self.assertEqual(new_revision.message, 'created from docker image')
        self.assertEqual(new_revision.repo, new_repo)
        self.assertEqual(new_ref.name, 'master')
        self.assertEqual(new_ref.repository, new_repo)
        self.assertEqual(new_version.worker.name, 'new_gen_classifier')
        self.assertEqual(new_version.worker.slug, 'new_gen_classifier')
        self.assertEqual(new_version.worker.public, False)
        self.assertEqual(new_version.worker.repository, new_repo)
        self.assertEqual(new_version.configuration, {'key': 'value'})
        self.assertEqual(new_version.docker_image_id, None)
        self.assertEqual(new_version.docker_image_iid, 'docker_image_42')
        self.assertEqual(new_version.gpu_usage, FeatureUsage.Disabled)
        self.assertEqual(new_version.model_usage, FeatureUsage.Disabled)
        self.assertEqual(new_worker_type.slug, 'docker')
        self.assertEqual(new_worker_type.display_name, 'Docker')
        # User is granted an admin role on the repository
        self.assertListEqual(list(new_repo.memberships.values_list("user", "level")), [
            (self.user.id, Role.Admin.value)
        ])
