from uuid import uuid4

from arkindex.documents.models import (
    Classification,
    Element,
    Entity,
    EntityType,
    MetaData,
    MLClass,
    Transcription,
    TranscriptionEntity,
)
from arkindex.ponos.models import Artifact
from arkindex.process.models import CorpusWorkerVersion, ProcessMode, Repository, WorkerRun
from arkindex.project.tests import FixtureTestCase


class TestManagers(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = Repository.objects.get(url='http://my_repo.fake/workers/worker')
        cls.revision = cls.repo.revisions.first()
        cls.artifact = Artifact.objects.get()
        # The fixtures have two worker versions, only one of them is used in existing elements

        cls.worker = cls.repo.workers.get(slug='reco')
        cls.worker_version = cls.worker.versions.get()
        cls.worker_configuration = cls.worker.configurations.create(name="test config")
        cls.model = cls.worker.models.create(name="test model")
        cls.model_version = cls.model.versions.create()

        cls.worker_run_1 = cls.worker_version.worker_runs.create(
            process=cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user),
        )
        cls.worker_run_2 = cls.worker_version.worker_runs.create(
            process=cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user),
            configuration=cls.worker_configuration,
        )
        cls.worker_run_3 = cls.worker_version.worker_runs.create(
            process=cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user),
            model_version=cls.model_version,
        )
        cls.worker_run_4 = cls.worker_version.worker_runs.create(
            process=cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user),
            model_version=cls.model_version,
            configuration=cls.worker_configuration,
        )

    def test_corpus_worker_version_rebuild(self):
        self.assertTrue(Element.objects.exists())
        Element.objects.update(
            worker_version=self.worker_version,
            worker_run=self.worker_run_1,
        )
        self.assertTrue(Transcription.objects.exists())
        Transcription.objects.update(
            worker_version=self.worker_version,
            worker_run=self.worker_run_1,
        )
        self.assertTrue(MetaData.objects.exists())
        Transcription.objects.update(
            worker_version=self.worker_version,
            worker_run=self.worker_run_2,
        )
        entity = Entity.objects.create(
            corpus=self.corpus,
            type=EntityType.objects.create(corpus=self.corpus, name="a"),
            worker_version=self.worker_version,
            worker_run=self.worker_run_3,
        )
        TranscriptionEntity.objects.create(
            entity=entity,
            transcription=Transcription.objects.first(),
            offset=42,
            length=1337,
            worker_version=self.worker_version,
            worker_run=self.worker_run_4,
        )
        # Test items having only a worker version set for compatibility
        Classification.objects.create(
            element=Element.objects.first(),
            ml_class=MLClass.objects.create(corpus=self.corpus, name="b"),
            worker_version=self.worker_version,
        )

        self.assertFalse(self.corpus.worker_version_cache.exists())
        CorpusWorkerVersion.objects.rebuild()
        self.assertCountEqual(
            [
                (c.worker_version_id, c.model_version_id, c.worker_configuration_id)
                for c in self.corpus.worker_version_cache.all()
            ],
            [
                (self.worker_version.id, None, None),
                (self.worker_version.id, self.model_version.id, None),
                (self.worker_version.id, None, self.worker_configuration.id),
                (self.worker_version.id, self.model_version.id, self.worker_configuration.id),
            ]
        )

    def test_worker_run_in_use(self):
        """
        Ensure the in_use method iterates over all related items
        """
        worker_run_id = uuid4()
        with self.assertExactQueries('worker_run_in_use.sql', params={'id': worker_run_id}):
            self.assertFalse(WorkerRun.objects.filter(id=worker_run_id).in_use())
