from unittest import TestCase
from unittest.mock import call, patch

from arkindex.project.openapi import check_tags, operation_id_pascal_case, request_body_x_name


@patch('arkindex.project.openapi.hooks.warn')
class TestSpectacularHooks(TestCase):

    def test_operation_id_pascal_case(self, warn_mock):
        """
        Test the operation_id_pascal_case post-processing hook that should
        turn camelCase operation IDs generated by Spectacular into PascalCase
        and warn about non-alphabetic IDs, as duplicate IDs can be resolved
        by Spectacular with numeric suffixes.
        """
        schema = operation_id_pascal_case({
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'retrieveA',
                    },
                    'post': {
                        'operationId': 'CreateA',
                    },
                    'put': {
                        'operationId': 'createA2',
                    }
                }
            }
        })
        self.assertDictEqual(schema, {
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'RetrieveA',
                    },
                    'post': {
                        'operationId': 'CreateA',
                    },
                    'put': {
                        'operationId': 'CreateA2',
                    }
                }
            }
        })
        self.assertEqual(warn_mock.call_count, 1)
        self.assertEqual(warn_mock.call_args, call(
            'Non-alphabetic characters found in operation ID createA2'
        ))

    @patch('arkindex.project.openapi.hooks.spectacular_settings.TAGS', [{'name': 'a'}])
    def test_check_tags(self, warn_mock):
        """
        Test the check_tags post-processing hook that should ensure each
        endpoint has at least one tag and only uses tags defined in
        arkindex.project.settings.
        """
        check_tags({
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'RetrieveA',
                    },
                    'post': {
                        'operationId': 'CreateA',
                        'tags': ['a'],
                    },
                    'delete': {
                        'operationId': 'DestroyA',
                        'tags': ['a', 'b'],
                    }
                }
            }
        })
        self.assertListEqual(warn_mock.call_args_list, [
            call('Endpoint RetrieveA does not have any assigned tags'),
            call("Endpoint DestroyA uses tags not defined in the project settings: {'b'}")
        ])

    def test_request_body_x_name(self, warn_mock):
        """
        Test the request_body_x_name post-processing hook adds x-name: body to all request bodies
        """
        schema = request_body_x_name({
            'paths': {
                '/a': {
                    'post': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            }
                        }
                    },
                    'delete': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            }
                        }
                    }
                }
            }
        })
        self.assertDictEqual(schema, {
            'paths': {
                '/a': {
                    'post': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            },
                            'x-name': 'body'
                        }
                    },
                    'delete': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            },
                            'x-name': 'body'
                        }
                    }
                }
            }
        })
