import copy
from textwrap import dedent
from uuid import UUID

from django.db import transaction
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response

from arkindex.documents.models import Corpus, Element
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, TrainingModelMixin
from arkindex.project.pagination import CountCursorPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.training.models import (
    Dataset,
    DatasetElement,
    DatasetState,
    MetricValue,
    Model,
    ModelVersion,
    ModelVersionState,
)
from arkindex.training.serializers import (
    CreateModelErrorResponseSerializer,
    DatasetElementInfoSerializer,
    DatasetElementSerializer,
    DatasetSerializer,
    ElementDatasetSerializer,
    MetricValueBulkSerializer,
    MetricValueCreateSerializer,
    ModelCompatibleWorkerSerializer,
    ModelSerializer,
    ModelVersionSerializer,
    ModelVersionValidateSerializer,
    SelectionDatasetElementSerializer,
)
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListModelVersions',
        description=(
            'List available versions of a Machine Learning model.\n\n'
            'Requires a **guest** access to the model, for versions that are available and have a tag,'
            'and a **contributor** access otherwise.'
        ),
    ),
    post=extend_schema(
        operation_id='CreateModelVersion',
        description=(
            'Create a new version for a Machine Learning model.\n\n'
            'The response includes an S3 URL that you can use to upload the archive containing the model.\n\n'
            'Once the archive is successfully uploaded, you can set the size and hash attributes using '
            '`PartialUpdateModelVersion`. The state of the model version will then be updated from `created` '
            'to `available`, meaning it can be used in a process.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: ModelVersionSerializer, 400: None, 403: None
        },
    ),
)
class ModelVersionsList(ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelVersionSerializer

    def get_queryset(self):
        self.model = self.get_object()
        qs = self.model.versions.order_by('-created')
        if self.access_level < Role.Contributor.value:
            qs = qs.filter(tag__isnull=False, state=ModelVersionState.Available)
        return qs

    def get_object(self):
        model = get_object_or_404(Model.objects.using('default').all(), id=self.kwargs['pk'])
        self.check_object_permissions(self.request, model)
        return model

    def check_object_permissions(self, request, model):
        self.access_level = get_max_level(self.request.user, model)
        needed_level, error_msg = (
            Role.Guest.value,
            'You need a Guest access to list versions of a model.',
        )
        if request.method not in permissions.SAFE_METHODS:
            needed_level, error_msg = (
                Role.Contributor.value,
                'You need a Contributor access to the model to create a new version.',
            )
        if not self.access_level or self.access_level < needed_level:
            raise PermissionDenied(detail=error_msg)
        return super().check_object_permissions(request, model)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update({
            'model': self.model,
            'is_contributor': self.access_level and self.access_level >= Role.Contributor.value,
        })
        return context

    def create(self, request, *args, **kwargs):
        self.model = self.get_object()
        return super().create(request, *args, **kwargs)


@extend_schema(tags=['training'])
@extend_schema_view(
    put=extend_schema(
        description=(
            'Update a version of a Machine Learning model.\n'
            'If the model version is updated with the hash of another available model version, '
            'an HTTP error 400 is returned containing data of the available model.\n\n'
            'Requires a **contributor** access to the model.'
        ),
    ),
    patch=extend_schema(
        description='Partially update a version of a Machine Learning model.\n\nRequires a **contributor** access to the model.'
    ),
    delete=extend_schema(
        description='Delete a model version.\n\nRequires an **admin** access on the related model.'
    )
)
class ModelVersionsRetrieve(RetrieveUpdateDestroyAPIView):
    """Retrieve a version of a Machine Learning model.

    Requires a **guest** access to the model, for versions that are available and have a tag,
    and a **contributor** access otherwise.
    """
    permission_classes = (IsVerified, )
    serializer_class = ModelVersionSerializer
    queryset = ModelVersion.objects.select_related("model")

    def get_serializer_context(self):
        context = super().get_serializer_context()
        access_level = getattr(self, 'access_level', None)
        context.update({
            'is_contributor': access_level is not None and access_level >= Role.Contributor.value,
        })
        return context

    def check_object_permissions(self, request, model_version):
        self.access_level = get_max_level(self.request.user, model_version.model)

        if request.method in permissions.SAFE_METHODS:
            # A Contributor access is required to retrieve an untagged or non available version
            if model_version.state != ModelVersionState.Available or model_version.tag is None:
                needed_level, error_msg = (
                    Role.Contributor.value,
                    'You need a Contributor access to the model to retrieve this version.',
                )
            else:
                needed_level, error_msg = (
                    Role.Guest.value,
                    'You need a Guest access to the model to retrieve this version.',
                )
        elif request.method.lower() == 'delete':
            needed_level, error_msg = (
                Role.Admin.value,
                'You need an Admin access to the model to destroy this version.',
            )
        else:
            needed_level, error_msg = (
                Role.Contributor.value,
                'You need a Contributor access to the model to update this version.',
            )

        if not self.access_level or self.access_level < needed_level:
            raise PermissionDenied(detail=error_msg)
        return super().check_object_permissions(request, model_version)


class ValidateModelVersion(TrainingModelMixin, GenericAPIView):
    """
    Checks the archive of a model version has correctly been uploaded to S3.
    If the verification succeeds, archive's attributes are updated and the state is set to `available`.
    If the verification fails, the state is set to `erroneous`.
    Once available, model versions cannot be validated anymore.
    If the hash conflicts with another existing model version, an HTTP error 409 is returned
    with the existing model version it the response body.

    Requires a **contributor** access to the model.
    """
    serializer_class = ModelVersionValidateSerializer
    queryset = ModelVersion.objects.select_related('model')

    def check_object_permissions(self, request, model_version):
        if not self.has_write_access(model_version.model):
            raise PermissionDenied(detail='You need a Contributor access to the model to validate this version.')
        if model_version.state == ModelVersionState.Available:
            raise PermissionDenied(detail='This model version is already marked as available.')
        return super().check_object_permissions(request, model_version)

    @extend_schema(
        operation_id='ValidateModelVersion',
        responses={
            200: ModelVersionSerializer,
            400: None,
            403: None,
            409: ModelVersionSerializer,
        },
        tags=['training'],
    )
    def post(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)

        # Ensure hash unicity among versions of a model
        existing_model_version = (
            instance.model.versions
            .filter(
                state=ModelVersionState.Available,
                hash=serializer.validated_data['hash'],
            ).first()
        )
        if existing_model_version:
            # Set the current model version as erroneous and return the available one
            instance.state = ModelVersionState.Error
            instance.save(update_fields=["state"])
            return Response(
                ModelVersionSerializer(
                    existing_model_version,
                    context={'is_contributor': True, 'model': instance},
                ).data,
                status=status.HTTP_409_CONFLICT,
            )

        for attr, value in serializer.validated_data.items():
            setattr(instance, attr, value)

        # Perform archive's validation
        try:
            instance.perform_check(save=False, raise_exc=True)
            instance.check_hash(save=False, raise_exc=True)
        except (AssertionError, ValueError) as e:
            # Set the model version as erroneous
            instance.save(update_fields=["state"])
            raise ValidationError({'detail': [str(e)]})

        instance.save()

        return Response(
            ModelVersionSerializer(instance).data,
            status=status.HTTP_201_CREATED,
        )


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListModels',
        description=(
            'List available Machine Learning models.'
        ),
        responses={
            200: ModelSerializer,
        },
        parameters=[
            OpenApiParameter(
                'name',
                description='Filter models whose name contains the given string (case insensitive)',
                required=False,
            ),
            OpenApiParameter(
                'compatible_worker',
                description='Only return models that are compatible with this worker.',
                type=UUID,
                required=False,
            ),
        ]
    ),
    post=extend_schema(
        description=(
            'Create a new Machine Learning model.'
        ),
        responses={
            200: ModelSerializer,
            400: CreateModelErrorResponseSerializer,
            403: None
        },
    ),
)
class ModelsList(TrainingModelMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelSerializer

    def get_queryset(self):
        filters = Q()
        if 'name' in self.request.query_params:
            filters &= Q(name__icontains=self.request.query_params['name'])
        if 'compatible_worker' in self.request.query_params:
            data = self.request.query_params['compatible_worker']
            try:
                data = UUID(data)
            except (TypeError, ValueError):
                raise serializers.ValidationError({'compatible_worker': [f'{data} is not a valid UUID.']})
            filters &= Q(compatible_workers__id=data)

        # Use the default database to prevent a stale read when a model has just been created
        return self.readable_models.using('default').filter(filters).order_by('name')

    def perform_create(self, serializer):
        model_name = serializer.validated_data.get('name')
        existing_model = Model.objects.using('default').filter(name=model_name).first()
        if existing_model:
            if self.has_read_access(existing_model):
                raise ValidationError({
                    'id': str(existing_model.id),
                    'name': 'A model with this name already exists',
                })
            else:
                raise PermissionDenied()
        return serializer.save()


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        description=dedent(
            """
            Retrieve a Machine Learning model.

            Requires a **guest** access to the model.
            """
        )
    ),
    put=extend_schema(
        description=dedent(
            """
            Update a Machine Learning model.

            Requires a **contributor** access to the model.
            """
        )
    ),
    patch=extend_schema(
        description=dedent(
            """
            Partially update a Machine Learning model.

            Requires a **contributor** access to the model.
            """
        )
    ),
)
class ModelRetrieve(TrainingModelMixin, RetrieveUpdateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelSerializer

    def get_queryset(self):
        return self.readable_models

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method in permissions.SAFE_METHODS:
            return
        if not self.has_write_access(obj):
            raise PermissionDenied(detail='You do not have a contributor access to this model.')


@extend_schema(tags=['training'], request=None)
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateModelCompatibleWorker',
        description=dedent(
            """
            Mark a worker as being compatible with a model.

            Requires **contributor** access on both the model and the worker.
            """
        ),
    ),
    delete=extend_schema(
        operation_id='DestroyModelCompatibleWorker',
        description=dedent(
            """
            Remove a worker from the list of compatible workers of a model.

            Requires **contributor** access on both the model and the worker.
            """
        ),
    ),
)
class ModelCompatibleWorkerManage(CreateAPIView, DestroyAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelCompatibleWorkerSerializer
    queryset = Model.compatible_workers.through.objects.all()

    def get_serializer(self, *args, **kwargs):
        # get_serializer may be called with no arguments just to get the serializer itself,
        # for example for the OpenAPI schema.
        if 'data' in kwargs:
            # It might be None, so we default it to an empty dict
            if not kwargs['data']:
                kwargs['data'] = {}
            kwargs['data'].update(**self.kwargs)

        return super().get_serializer(*args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        # DestroyModelMixin normally does not use any serializer, so we rewrite it
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        get_object_or_404(self.get_queryset(), **serializer.validated_data).delete()
        headers = self.get_success_headers(serializer.data)
        return Response(status=status.HTTP_204_NO_CONTENT, headers=headers)


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='DownloadModelVersion',
        description=(
            'Retrieve a download url for a model_version. A secret token is required to have access to the model version.'
            'This endpoint does **not** require to be logged in.'
        ),
        responses={
            302: serializers.Serializer,
        },
        parameters=[
            OpenApiParameter(
                'token',
                description='SHA256 hash used for authentication and check access to the model version.',
                required=True,
            ),
        ]
    ),
)
class ModelVersionDownload(RetrieveAPIView):
    queryset = ModelVersion.objects.all()

    def check_object_permissions(self, request, model_version):
        if model_version.state != ModelVersionState.Available:
            raise PermissionDenied(detail='This model version is not available yet.')
        token = model_version.build_authentication_token_hash()
        if token != request.query_params['token']:
            raise PermissionDenied(detail='Wrong token.')

    def get(self, request, *args, **kwargs):
        return Response(status=status.HTTP_302_FOUND, headers={'Location': self.get_object().s3_url})


@extend_schema(tags=['training'])
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateMetricValue',
        description=(
            'Create a metric on a model version training.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: MetricValueCreateSerializer, 400: None, 403: None
        },
    )
)
class MetricValueCreate(CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = MetricValueCreateSerializer
    queryset = MetricValue.objects.none()


@extend_schema(tags=['training'])
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateMetricValues',
        description=(
            'Bulk create metrics on a model version training, optionally on a specified step.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: MetricValueCreateSerializer, 400: None, 403: None
        },
    )
)
class MetricValueBulkCreate(CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = MetricValueBulkSerializer
    queryset = MetricValue.objects.none()


@extend_schema_view(
    get=extend_schema(
        operation_id='ListCorpusDatasets',
        description=dedent(
            """
            List all datasets in a corpus.

            Requires a **guest** access to the corpus.
            """
        ),
    ),
    post=extend_schema(
        operation_id='CreateDataset',
        description=dedent(
            """
            Create a new dataset in a corpus.

            Requires a **contributor** access to the corpus.
            """
        )
    ),
)
@extend_schema(tags=['datasets'])
class CorpusDataset(CorpusACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = DatasetSerializer
    queryset = Dataset.objects.none()

    def get_queryset(self):
        return Dataset.objects \
            .select_related('creator') \
            .filter(corpus=self.corpus) \
            .order_by('name')

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method not in permissions.SAFE_METHODS:
            role = Role.Contributor
        return self.get_corpus(self.kwargs['pk'], role=role)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.kwargs:
            return context
        context['corpus'] = self.corpus
        return context


@extend_schema_view(
    get=extend_schema(
        description=dedent(
            """
            Retrieve a dataset.

            Requires a **guest** access to the dataset's corpus.
            """
        )
    ),
    put=extend_schema(
        description=dedent(
            """
            Update a dataset.

            Requires a **contributor** access to the dataset's corpus.
            """
        ),
    ),
    patch=extend_schema(
        description=dedent(
            """
            Partially update a dataset.

            Requires a **contributor** access to the dataset's corpus.
            """
        )
    ),
    delete=extend_schema(
        description=dedent(
            """
            Delete a dataset. Only datasets in an `open` state can be deleted.

            Requires an **admin** access to the dataset's corpus.
            """
        )
    )
)
@extend_schema(tags=['datasets'])
class DatasetUpdate(ACLMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = (IsVerified,)
    serializer_class = DatasetSerializer

    def get_queryset(self):
        queryset = Dataset.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))
        return queryset.select_related('corpus', 'creator')

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if request.method in permissions.SAFE_METHODS:
            return
        role = Role.Contributor
        if request.method == 'DELETE':
            role = Role.Admin
            if obj.state != DatasetState.Open:
                raise PermissionDenied(detail='Only datasets in an open state can be deleted.')
        if not self.has_access(obj.corpus, role.value):
            raise PermissionDenied(detail=f'You do not have {str(role).lower()} access to corpus {obj.corpus.name}.')

    def perform_destroy(self, dataset):
        dataset.dataset_elements.all().delete()
        super().perform_destroy(dataset)


class DatasetElementCursorPagination(CountCursorPagination):
    # The same element can be in multiple sets at once within the dataset,
    # so we fall back on the DatasetElement's ID
    ordering = ('element_id', 'id')


@extend_schema(tags=['datasets'])
class DatasetElements(CorpusACLMixin, ListAPIView):
    """
    List all elements in a dataset.\n\n
    Requires a **guest** access to the dataset corpus.
    """
    permission_classes = (IsVerified, )
    queryset = DatasetElement.objects.none()
    serializer_class = DatasetElementSerializer
    pagination_class = DatasetElementCursorPagination

    def get_queryset(self):
        dataset = get_object_or_404(
            Dataset.objects.select_related('corpus'),
            id=self.kwargs['pk'],
        )
        if not self.has_read_access(dataset.corpus):
            raise PermissionDenied(detail='You do not have access to the corpus of this dataset')
        return (
            dataset.dataset_elements
            .prefetch_related('element')
            .select_related('element__type', 'element__corpus', 'element__image__server')
        )


@extend_schema_view(
    delete=extend_schema(
        operation_id='DestroyDatasetElement',
        parameters=[
            OpenApiParameter(
                'set',
                type=str,
                description='Name of the set from which to remove the element.',
                required=True,
            )
        ],
        tags=['datasets']
    )
)
class DatasetElementDestroy(CorpusACLMixin, DestroyAPIView):
    """
    Remove an element from a dataset.

    Elements can only be removed from **open** datasets.

    Requires a **contributor** access to the dataset corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = DatasetElementInfoSerializer
    lookup_url_kwarg = 'element'

    def destroy(self, request, *args, **kwargs):
        if not self.request.query_params.get('set'):
            raise ValidationError({'set': ['This field is required.']})
        dataset_element = get_object_or_404(
            DatasetElement.objects.select_related('dataset__corpus'),
            dataset_id=self.kwargs['dataset'],
            element_id=self.kwargs['element'],
            set=self.request.query_params.get('set')
        )
        if dataset_element.dataset.state != DatasetState.Open:
            raise ValidationError({'dataset': ['Elements can only be removed from open Datasets.']})
        if not self.has_write_access(dataset_element.dataset.corpus):
            raise PermissionDenied(detail='You need a Contributor access to the dataset to perform this action.')
        dataset_element.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(tags=['datasets'])
@extend_schema_view(
    post=extend_schema(
        operation_id='CreateDatasetElementsSelection',
        responses={
            204: None,
            400: None,
            403: None,
        },
        parameters=[
            OpenApiParameter(
                'id',
                type=UUID,
                location=OpenApiParameter.PATH,
                description='ID of the corpus containing the selected elements.',
                required=True,
            )
        ],
    )
)
class CreateDatasetElementsSelection(CorpusACLMixin, CreateAPIView):
    """
    Add elements from a selection to a dataset.

    Requires a **contributor** access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = SelectionDatasetElementSerializer

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, corpus):
        if not self.has_write_access(corpus):
            raise PermissionDenied(detail="You need a Contributor access to the corpus to perform this action.")
        super().check_object_permissions(request, corpus)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus'] = self.corpus
        return context

    @cached_property
    def corpus(self):
        return self.get_object()

    def create(self, request, *args, **kwargs):
        super().create(request, *args, **kwargs)
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(
    tags=['datasets'],
    parameters=[
        OpenApiParameter(
            'id',
            type=UUID,
            location=OpenApiParameter.PATH,
            description='ID of the element.',
            required=True,
        )
    ],
)
class ElementDatasets(CorpusACLMixin, ListAPIView):
    """
    List all datasets containing a specific element.

    Requires a **guest** access to the element's corpus.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = ElementDatasetSerializer

    @cached_property
    def element(self):
        return get_object_or_404(
            Element.objects.only('id'),
            id=self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user)
        )

    def get_queryset(self):
        return (
            self.element.dataset_elements.all()
            .select_related('dataset__creator')
            .order_by('dataset__name', 'set', 'dataset_id')
        )


@extend_schema_view(
    post=extend_schema(
        tags=['datasets'],
        request=None,
        operation_id='CloneDataset',
        parameters=[
            OpenApiParameter(
                'id',
                type=UUID,
                location=OpenApiParameter.PATH,
                description='ID of the dataset to copy.',
                required=True,
            )
        ]
    ),
)
class DatasetClone(CorpusACLMixin, CreateAPIView):
    """
    Clone a dataset and return its copy. The new dataset's state is set to Open.

    Requires a **contributor** access to the dataset's corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = DatasetSerializer

    def get_queryset(self):
        return Dataset.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, dataset):
        if not self.has_write_access(dataset.corpus):
            raise PermissionDenied(detail="You need a Contributor access to the dataset's corpus.")
        return super().check_object_permissions(request, dataset)

    @transaction.atomic
    def create(self, request, *args, **kwargs):
        dataset = self.get_object()

        # Dataset names must be unique
        prefix = 'Clone of '
        base_clone_name = clone_name = dataset.name if dataset.name.startswith(prefix) else f'{prefix}{dataset.name}'
        existing = (
            Dataset.objects
            .filter(
                corpus_id=dataset.corpus_id,
                name__startswith=clone_name
            )
            .values_list('name', flat=True)
        )
        index = 1
        while clone_name in existing:
            clone_name = f'{base_clone_name} {index}'
            index += 1

        clone = copy.copy(dataset)
        # Make Django think it is a new dataset that it should insert
        clone.id = None
        clone.name = clone_name[:100]
        clone.state = DatasetState.Open
        clone.creator = request.user
        clone.save()

        # Associate all elements to the clone
        DatasetElement.objects.bulk_create([
            DatasetElement(element_id=elt_id, dataset_id=clone.id, set=set_name)
            for elt_id, set_name in dataset.dataset_elements.values_list('element_id', 'set')
        ])
        return Response(
            DatasetSerializer(clone).data,
            status=status.HTTP_201_CREATED,
        )
